/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "stdlib.h"
#include "compute_reduce_region.h"
#include "atom.h"
#include "update.h"
#include "modify.h"
#include "domain.h"
#include "group.h"
#include "region.h"
#include "fix.h"
#include "input.h"
#include "variable.h"
#include "memory.h"
#include "error.h"

using namespace LAMMPS_NS;

enum{SUM,MINN,MAXX};
enum{X,V,F,COMPUTE,FIX,VARIABLE};
enum{GLOBAL,PERATOM,LOCAL};

#define INVOKED_VECTOR 2
#define INVOKED_ARRAY 4
#define INVOKED_PERATOM 8
#define INVOKED_LOCAL 16

#define BIG 1.0e20

/* ---------------------------------------------------------------------- */

ComputeReduceRegion::ComputeReduceRegion(LAMMPS *lmp, int narg, char **arg) :
  ComputeReduce(lmp, narg, arg) {}

/* ----------------------------------------------------------------------
   calculate reduced value for one input M and return it
   if flag = -1:
     sum/min/max/ave all values in vector
     for per-atom quantities, limit to atoms in group and region
     if mode = MIN or MAX, also set index to which vector value wins
   if flag >= 0: simply return vector[flag]
------------------------------------------------------------------------- */

double ComputeReduceRegion::compute_one(int m, int flag)
{
  int i;

  Region *region = domain->regions[iregion];

  // invoke the appropriate attribute,compute,fix,variable
  // compute scalar quantity by summing over atom scalars
  // only include atoms in group

  index = -1;
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  int n = value2index[m];
  int j = argindex[m];

  double one;
  if (mode == SUM) one = 0.0;
  else if (mode == MINN) one = BIG;
  else if (mode == MAXX) one = -BIG;

  if (which[m] == X) {
    if (flag < 0) {
      for (i = 0; i < nlocal; i++)
	if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	  combine(one,x[i][j],i);
    } else one = x[flag][j];
  } else if (which[m] == V) {
    double **v = atom->v;
    if (flag < 0) {
      for (i = 0; i < nlocal; i++)
	if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	  combine(one,v[i][j],i);
    } else one = v[flag][j];
  } else if (which[m] == F) {
    double **f = atom->f;
    if (flag < 0) {
      for (i = 0; i < nlocal; i++)
	if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	  combine(one,f[i][j],i);
    } else one = f[flag][j];
    
  // invoke compute if not previously invoked

  } else if (which[m] == COMPUTE) {
    Compute *compute = modify->compute[n];
    
    if (flavor[m] == GLOBAL) {
      if (j == 0) {
	if (!(compute->invoked_flag & INVOKED_VECTOR)) {
	  compute->compute_vector();
	  compute->invoked_flag |= INVOKED_VECTOR;
	}
	double *compute_vector = compute->vector;
	int n = compute->size_vector;
	if (flag < 0) 
	  for (i = 0; i < n; i++)
	    combine(one,compute_vector[i],i);
	else one = compute_vector[flag];
      } else {
	if (!(compute->invoked_flag & INVOKED_ARRAY)) {
	  compute->compute_array();
	  compute->invoked_flag |= INVOKED_ARRAY;
	}
	double **compute_array = compute->array;
	int n = compute->size_array_rows;
	int jm1 = j - 1;
	if (flag < 0) 
	  for (i = 0; i < n; i++)
	    combine(one,compute_array[i][jm1],i);
	else one = compute_array[flag][jm1];
      }

    } else if (flavor[m] == PERATOM) {
      if (!(compute->invoked_flag & INVOKED_PERATOM)) {
	compute->compute_peratom();
	compute->invoked_flag |= INVOKED_PERATOM;
      }

      if (j == 0) {
	double *compute_vector = compute->vector_atom;
	int n = nlocal;
	if (flag < 0) {
	  for (i = 0; i < n; i++)
	    if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	      combine(one,compute_vector[i],i);
	} else one = compute_vector[flag];
      } else {
	double **compute_array = compute->array_atom;
	int n = nlocal;
	int jm1 = j - 1;
	if (flag < 0) {
	  for (i = 0; i < n; i++)
	    if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	      combine(one,compute_array[i][jm1],i);
	} else one = compute_array[flag][jm1];
      }

    } else if (flavor[m] == LOCAL) {
      if (!(compute->invoked_flag & INVOKED_LOCAL)) {
	compute->compute_local();
	compute->invoked_flag |= INVOKED_LOCAL;
      }

      if (j == 0) {
	double *compute_vector = compute->vector_local;
	int n = compute->size_local_rows;
	if (flag < 0)
	  for (i = 0; i < n; i++)
	    combine(one,compute_vector[i],i);
	else one = compute_vector[flag];
      } else {
	double **compute_array = compute->array_local;
	int n = compute->size_local_rows;
	int jm1 = j - 1;
	if (flag < 0)
	  for (i = 0; i < n; i++)
	    combine(one,compute_array[i][jm1],i);
	else one = compute_array[flag][jm1];
      }
    }

  // check if fix frequency is a match

  } else if (which[m] == FIX) {
    if (update->ntimestep % modify->fix[n]->peratom_freq)
      error->all("Fix used in compute reduce not computed at compatible time");
    Fix *fix = modify->fix[n];

    if (flavor[m] == GLOBAL) {
      if (j == 0) {
	int n = fix->size_vector;
	if (flag < 0)
	  for (i = 0; i < n; i++)
	    combine(one,fix->compute_vector(i),i);
	else one = fix->compute_vector(flag);
      } else {
	int n = fix->size_array_rows;
	int jm1 = j - 1;
	if (flag < 0)
	  for (i = 0; i < nlocal; i++)
	    combine(one,fix->compute_array(i,jm1),i);
	else one = fix->compute_array(flag,jm1);
      }

    } else if (flavor[m] == PERATOM) {
      if (j == 0) {
	double *fix_vector = fix->vector_atom;
	int n = nlocal;
	if (flag < 0) {
	  for (i = 0; i < n; i++)
	    if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	      combine(one,fix_vector[i],i);
	} else one = fix_vector[flag];
      } else {
	double **fix_array = fix->array_atom;
	int n = nlocal;
	int jm1 = j - 1;
	if (flag < 0) {
	  for (i = 0; i < nlocal; i++)
	    if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	      combine(one,fix_array[i][jm1],i);
	} else one = fix_array[flag][jm1];
      }

    } else if (flavor[m] == LOCAL) {
      if (j == 0) {
	double *fix_vector = fix->vector_local;
	int n = fix->size_local_rows;
	if (flag < 0)
	  for (i = 0; i < n; i++)
	    combine(one,fix_vector[i],i);
	else one = fix_vector[flag];
      } else {
	double **fix_array = fix->array_local;
	int n = fix->size_local_rows;
	int jm1 = j - 1;
	if (flag < 0)
	  for (i = 0; i < n; i++)
	    combine(one,fix_array[i][jm1],i);
	else one = fix_array[flag][jm1];
      }
    }
    
  // evaluate atom-style variable

  } else if (which[m] == VARIABLE) {
    if (nlocal > maxatom) {
      maxatom = atom->nmax;
      memory->sfree(varatom);
      varatom =	(double *) 
	memory->smalloc(maxatom*sizeof(double),"reduce/region:varatom");
    }

    input->variable->compute_atom(n,igroup,varatom,1,0);
    if (flag < 0) {
      for (i = 0; i < nlocal; i++)
	if (mask[i] & groupbit && region->match(x[i][0],x[i][1],x[i][2]))
	  combine(one,varatom[i],i);
    } else one = varatom[flag];
  }

  return one;
}

/* ---------------------------------------------------------------------- */

double ComputeReduceRegion::count(int m)
{
  int n = value2index[m];
  int j = argindex[m];

  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  if (which[m] == X || which[m] == V || which[m] == F)
    return group->count(igroup,iregion);
  else if (which[m] == COMPUTE) {
    Compute *compute = modify->compute[n];
    if (flavor[m] == GLOBAL) {
      if (j == 0) return(1.0*compute->size_vector);
      else return(1.0*compute->size_array_rows);
    } else if (flavor[m] == PERATOM) {
      return group->count(igroup,iregion);
    } else if (flavor[m] == LOCAL) {
      double ncount = compute->size_local_rows;
      double ncountall;
      MPI_Allreduce(&ncount,&ncountall,1,MPI_DOUBLE,MPI_SUM,world);
      return ncountall;
    }
  } else if (which[m] == FIX) {
    Fix *fix = modify->fix[n];
    if (flavor[m] == GLOBAL) {
      if (j == 0) return(1.0*fix->size_vector);
      else return(1.0*fix->size_array_rows);
    } else if (flavor[m] == PERATOM) {
      return group->count(igroup,iregion);
    } else if (flavor[m] == LOCAL) {
      double ncount = fix->size_local_rows;
      double ncountall;
      MPI_Allreduce(&ncount,&ncountall,1,MPI_DOUBLE,MPI_SUM,world);
      return ncountall;
    }
  } else if (which[m] == VARIABLE)
    return group->count(igroup,iregion);

  return 0.0;
}
